/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "llm.h"

#include <sstream>
#include <nlohmann/json.hpp>

namespace tc = triton::client;

#define FAIL_IF_ERR(X, MSG)                                                    \
    {                                                                          \
        tc::Error err = (X);                                                   \
        if (!err.IsOk()) {                                                     \
            std::cerr << "error: " << (MSG) << ": " << err << std::endl;       \
            exit(1);                                                           \
        }                                                                      \
    }

namespace {

void ValidateShapeAndDatatype(const std::string &name,
                              std::shared_ptr<tc::InferResult> result) {
    std::vector<int64_t> shape;
    FAIL_IF_ERR(result->Shape(name, &shape),
                "unable to get shape for '" + name + "'");
    // Validate shape
    if ((shape.size() != 2) || shape[0] != 1 || shape[1] != 1) {
        std::cerr << "error: received incorrect shapes for '" << name << "'"
                  << std::endl;
        exit(1);
    }
    std::string datatype;
    FAIL_IF_ERR(result->Datatype(name, &datatype),
                "unable to get datatype for '" + name + "'");
    // Validate datatype
    if (datatype.compare("BYTES") != 0) {
        std::cerr << "error: received incorrect datatype for '" << name
                  << "': " << datatype << std::endl;
        exit(1);
    }
}

} // namespace

void LLM::llm_callback(tc::InferResult *result) {
    std::shared_ptr<tc::InferResult> result_ptr(result);
    std::lock_guard<std::mutex> lk(mutex_);
    result_list_.push_back(result_ptr);
    cv_.notify_all();
}

bool LLM::llm_load_model() {
    tc::Error err_loadmodel = loadClient_->LoadModel(modelName_);
    if (!err_loadmodel.IsOk()) {
        std::cout << __LINE__
                  << " err_loadmodel.Message(): " << err_loadmodel.Message()
                  << std::endl;
        return false;
    }

    modelInfoUpdate();
    return true;
}

namespace nl = nlohmann;
LLM::LLM() {
    bool verbose = true;
#if 1
    tc::Error err_create = tc::InferenceServerGrpcClient::Create(
        &loadClient_, grpcServerUrl_, verbose);
    if (!err_create.IsOk()) {
        std::cout << __LINE__
                  << " err_create.Message(): " << err_create.Message()
                  << std::endl;
    }
#else
    bool use_ssl = false;
    bool use_cached_channel = true;
    tc::SslOptions ssl_options = tc::SslOptions();
    tc::Error err_create = KyInfer::Client::CreateGrpcClient(
        &loadClient_, grpcServerUrl_, verbose, use_ssl, ssl_options,
        tc::KeepAliveOptions(), use_cached_channel);
#endif

    // FAIL_IF_ERR(
    //     KyInfer::Client::CreateGrpcClient(&loadClient_, httpServerUrl_,
    //     verbose), "unable to create http client");

    tc::Error err;
    int reTryTime = 25;
    do {
        err = loadClient_->LoadModel("config_bank");
        if (!err.IsOk()) {
            std::cout << __LINE__
                      << " LoadModel err.Message(): " << err.Message()
                      << std::endl;
            // return "";
            sleep(1);
            reTryTime--;
        }
    } while (!err.IsOk() && reTryTime > 0);

    modelName_ = GetConfigValue("llm", "llm");
    if (modelName_.size() < 1) {
        printf("line:%d model_name.size():%ld\n", __LINE__, modelName_.size());
        modelName_ = "Qwen-2.5-3b_1.0";
    }
    // loadClient_->LoadModel(modelName_);

    tc::Error err_create_grpc = tc::InferenceServerGrpcClient::Create(
        &grpcClient_, grpcServerUrl_, verbose);
    auto now = std::chrono::system_clock::now();
    auto value = now.time_since_epoch();
    auto microseconds =
        std::chrono::duration_cast<std::chrono::microseconds>(value).count();
    double seconds = microseconds / 1000000.0;
    request_id_ = std::to_string(seconds);
    printf("%d: request_id_:%s\n", __LINE__, request_id_.c_str());
    tc::Headers http_headers;
    uint32_t stream_timeout = 0;
    FAIL_IF_ERR(grpcClient_->StartStream(
                    // llm_callback,
                    std::bind(&LLM::llm_callback, this, std::placeholders::_1),
                    false /*ship_stats*/, stream_timeout, http_headers),
                "unable to establish a streaming connection to server");
    default_model_info_ = R"(
        {
            "vendor": "",
            "models": [{
                "name": "Qwen-2.5-3b_1.0",
                "capability": "llm",
                "subConfig": [],
                "debName": "kylin-qwen2.5-3b-gguf-model",
                "is3rdModel": "false",
                "maxTokens": 1024
            }]
        })";

    // modelInfo_ = GetConfigValue("llm_support", "llm");
    modelInfo_ = R"(
      {
          "vendor": "",
          "models": []
      })";
    // default n_predict_input_ = 1024;
    n_predict_input_ = 1024;
    modelInfoUpdate();
    printf("modelInfo_:%s\n", modelInfo_.c_str());
}

void LLM::modelInfoUpdate() {
    std::string model_info_list = GetConfigValue("model_info_list", "llm");

    nl::json j = nl::json::parse(modelInfo_);
    try {
        j["models"].clear();
        nl::json new_models = nl::json::parse(model_info_list)["models"];
        j["models"].insert(j["models"].end(), new_models.begin(),
                           new_models.end());

        for (auto &it_json : j["models"]) {
            // 获取模型配置
            inference::ModelConfigResponse config;
            tc::Error err_modelconfig =
                loadClient_->ModelConfig(&config, it_json["name"]);

            if (!err_modelconfig.IsOk()) {
                std::cout
                    << __LINE__
                    << " unable to get ModelConfig, err_modelconfig.Message(): "
                    << err_modelconfig.Message() << std::endl;
            } else {
                // 查找名为 'n_ctx' 的参数
                int n_ctx = 0;
                auto it = config.config().parameters().find("n_ctx");
                if (it != config.config().parameters().end()) {
                    std::cout << "n_ctx parameter value: "
                              << it->second.string_value() << std::endl;

                    n_ctx = std::stoi(it->second.string_value());
                } else {
                    std::cout << "n_ctx parameter not found." << std::endl;
                    n_ctx = 0;
                }
                it_json["maxTokens"] = n_ctx;

                auto it_n_predic =
                    config.config().parameters().find("n_predict_input");
                if (it_n_predic != config.config().parameters().end()) {
                    std::cout << "n_predict_input parameter value: "
                              << it_n_predic->second.string_value()
                              << std::endl;

                    n_predict_input_ =
                        std::stoi(it_n_predic->second.string_value());
                } else {
                    std::cout
                        << "n_predict_input parameter not found. set to 1024"
                        << std::endl;
                    n_predict_input_ = 1024;
                }
            }
        }
        modelInfo_ = j.dump(4);
        // 成功解析后的代码逻辑
    } catch (const nl::json::parse_error &e) {
        // 捕获解析错误
        std::cerr << "Parse error at byte " << e.byte << ": " << e.what()
                  << '\n';
        modelInfo_ = default_model_info_;
    } catch (const nl::json::exception &e) {
        // 捕获所有来自nlohmann::json库的异常
        std::cerr << "JSON exception: " << e.what() << '\n';
        modelInfo_ = default_model_info_;
    } catch (const std::exception &e) {
        // 捕获所有标准C++异常
        std::cerr << "Standard exception: " << e.what() << '\n';
        modelInfo_ = default_model_info_;
    }

    if (modelInfo_.size() < 1) {
        std::cerr << "modelInfo_.size() < 1 : " << modelInfo_.size() << '\n';
        modelInfo_ = default_model_info_;
    }
}

LLM::~LLM() {
    // TODO:
}

bool LLM::setLlmModelName(const std::string &modelName) {
    std::string change_cmd = "change_" + modelName;
    nl::json llm_json;
    if (modelInfo_.size() > 0)
        llm_json = nl::json::parse(modelInfo_);
    else
        return false;
    for (auto llm_obj : llm_json["models"]) {
        if ((modelName_ != modelName) && (modelName == llm_obj["name"])) {
            std::string change_result =
                GetConfigValue(change_cmd.c_str(), "llm");
            // printf("----%d %s
            // change_cmd:%s\n",__LINE__,__FUNCTION__,change_cmd.c_str());
            // printf("----%d %s
            // change_result:%s\n",__LINE__,__FUNCTION__,change_result.c_str());
            loadClient_->UnloadModel(modelName_);
            modelName_ = modelName;
            loadClient_->LoadModel(modelName_);
            modelInfoUpdate();
            return true;
        }
    }
    return false;
}

void LLM::setCompletionResultCallback(CompletionResultCallback callback) {
    // std::unique_lock<std::mutex> lock(sessionMutex_);

    completionCallbacks_ = std::move(callback);
}

void LLM::stopChat() {
    if (grpcClient_ == nullptr) {
        return;
    }
    try {
        std::cout << __LINE__ << "in stopChat" << std::endl;
        bool verbose = true;

        // tc::Error err_create =
        // tc::InferenceServerGrpcClient::Create(&grpcClient_, grpcServerUrl_,
        // verbose);
        struct chatConfig config;

        int lora_count = getLoraCount();
        std::vector<float> lora_scale_input_data(lora_count, 0.0f);
        setLoraScale(-1, lora_scale_input_data);

        config.request_type_data = 5;
        config.request_id = request_id_;
        config.chatPrompt = "";
        config.temperature_input = 0.0;
        config.temperature_input_data = 0.0;
        config.top_k_input_data = 0;
        config.top_p_input_data = 0.0;
        config.n_keep_input_data = 0;
        config.n_predict_input_data = 0;
        config.cache_prompt_input_data = true;
        config.stop_input_data = {"<|im_end|>"};
        config.lora_scale_input_data = lora_scale_input_data;
        config.stream_input_data = true;
        std::vector<tc::InferInput *> inputs;
        tc::Headers http_headers;
        uint32_t stream_timeout = 0;
        std::string ans;
        bool finalResponse = false;

        loadClient_->LoadModel(modelName_);
        chat_inputs(config, inputs);

        return;
    } catch (const std::exception &e) {
        // 捕获所有从 std::exception 派生的异常
        std::cerr << "Caught exception: " << e.what() << std::endl;
    }
}

std::string LLM::GetConfigValue(const char *key, const char *engine) {
    // GetConfigValue("change_ppocr", "ocr");
    tc::Error err = loadClient_->LoadModel("config_bank");
    std::string model_name = "config_bank";
    tc::InferOptions options(model_name);
    options.model_version_ = "";
    options.priority_ = 1;

    std::vector<std::string> input_key_data{std::string(key)};
    std::vector<std::string> input_engine_data{std::string(engine)};

    tc::InferInput *input0, *input1;
    std::vector<int64_t> shape{1};
    tc::InferInput::Create(&input0, "key", shape, "BYTES");
    tc::InferInput::Create(&input1, "engine", shape, "BYTES");
    std::shared_ptr<tc::InferInput> input0_ptr, input1_ptr;
    input0_ptr.reset(input0);
    input1_ptr.reset(input1);
    input0_ptr->AppendFromString(input_key_data);
    input1_ptr->AppendFromString(input_engine_data);
    std::vector<tc::InferInput *> inputs = {input0_ptr.get(), input1_ptr.get()};

    tc::InferRequestedOutput *output0;
    tc::Error err_out0 = tc::InferRequestedOutput::Create(&output0, "value");
    std::shared_ptr<tc::InferRequestedOutput> output0_ptr;
    output0_ptr.reset(output0);
    std::vector<const tc::InferRequestedOutput *> outputs = {output0_ptr.get()};

    tc::InferResult *results;
    tc::Headers http_headers;
    tc::Error err_infer =
        loadClient_->Infer(&results, options, inputs, outputs, http_headers);
    if (!err_infer.IsOk()) {
        std::cout << __LINE__ << " err_infer.Message(): " << err_infer.Message()
                  << std::endl;
        return "";
    }

    std::shared_ptr<tc::InferResult> results_ptr;
    results_ptr.reset(results);

    char *output0_data_ptr;
    size_t recv_output0_byte_size;

    std::vector<std::string> text_output_data;
    tc::Error err_paser = results_ptr->StringData("value", &text_output_data);
    if (!err_paser.IsOk()) {
        std::cout << __LINE__ << " err_paser.Message(): " << err_infer.Message()
                  << std::endl;
        return "";
    }
    return text_output_data[0];
}

int LLM::setLoraScale(int task_id, std::vector<float> &lora_scale_input_data) {
    // 检查task_id是否是有效索引
    if (task_id >= 0 &&
        task_id < static_cast<int>(lora_scale_input_data.size())) {
        // 填充所有元素为0
        std::fill(lora_scale_input_data.begin(), lora_scale_input_data.end(),
                  0.0f);

        // 设置指定索引的值为1
        lora_scale_input_data[task_id] = 1.0f;
    } else {
        // throw std::out_of_range("task_id is out of the vector's range.");
        return -1;
    }
    return 0;
}

int LLM::getLoraCount() {
    // 获取模型配置
    inference::ModelConfigResponse config;
    tc::Error err_modelconfig = loadClient_->ModelConfig(&config, modelName_);

    if (!err_modelconfig.IsOk()) {
        std::cout << __LINE__
                  << " unable to get ModelConfig, err_modelconfig.Message(): "
                  << err_modelconfig.Message() << std::endl;
        return -1;
    }

    // 查找名为 'lora' 的参数
    int semicolon_count = 0;
    auto it = config.config().parameters().find("lora");
    if (it != config.config().parameters().end()) {
        std::cout << "LoRA parameter value: " << it->second.string_value()
                  << std::endl;
        if (it->second.string_value().empty())
            return 0;

        std::istringstream stream(it->second.string_value());
        std::string segment;
        int count = 0;

        while (std::getline(stream, segment, ';')) {
            if (!segment.empty()) { // 忽略空段
                ++count;
            }
        }

        // 如果最后一个字符是分隔符，则原始字符串可能有一个额外的空段
        if (!it->second.string_value().empty() &&
            it->second.string_value().back() == ';') {
            ++count;
        }

        semicolon_count = count;
    } else {
        std::cout << "LoRA parameter not found." << std::endl;
        semicolon_count = 0;
    }

    return semicolon_count;
}

void LLM::chat_inputs(struct chatConfig config,
                      std::vector<tc::InferInput *> &inputs) {

    // input: request_type
    int32_t request_type_data = 0;
    tc::InferInput *request_type;
    std::vector<int64_t> request_type_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&request_type, "request_type",
                                       request_type_shape, "INT32"),
                "unable to create 'request_type'");
    std::shared_ptr<tc::InferInput> request_type_ptr(request_type);
    FAIL_IF_ERR(request_type_ptr->Reset(), "unable to reset 'request_type'");
    FAIL_IF_ERR(request_type_ptr->AppendRaw(
                    reinterpret_cast<uint8_t *>(&config.request_type_data),
                    sizeof(int32_t)),
                "unable to set data for 'request_type'");
    std::cout << "nlp_param->request_type: " << config.request_type_data
              << std::endl;

    // input: request_id
    std::vector<std::string> request_id_data(1);
    std::vector<int64_t> request_id_shape{1};
    request_id_data[0] = config.request_id;
    tc::InferInput *request_id;
    FAIL_IF_ERR(tc::InferInput::Create(&request_id, "request_id",
                                       request_id_shape, "BYTES"),
                "unable to create 'request_id'");
    std::shared_ptr<tc::InferInput> request_id_ptr;
    request_id_ptr.reset(request_id);
    request_id_ptr->SetBinaryData(true);
    FAIL_IF_ERR(request_id_ptr->AppendFromString(request_id_data),
                "unable to set data for request_id");
    printf("request_id_data[0]:%s\n", request_id_data[0].c_str());

    // input: text_input
    std::vector<std::string> text_input_data(1);
    std::vector<int64_t> text_input_shape{1};
    text_input_data[0] = config.chatPrompt;
    tc::InferInput *text_input;
    FAIL_IF_ERR(tc::InferInput::Create(&text_input, "text_input",
                                       text_input_shape, "BYTES"),
                "unable to get text_input");
    std::shared_ptr<tc::InferInput> text_input_ptr;
    text_input_ptr.reset(text_input);
    text_input_ptr->SetBinaryData(true);
    FAIL_IF_ERR(text_input_ptr->AppendFromString(text_input_data),
                "unable to set data for text_input");
    printf("text_input_data[0]:%s\n", text_input_data[0].c_str());

    // input: temperature_input
    // int32_t temperature_input_data = 0.8;
    int32_t temperature_input_data = 0.0;
    tc::InferInput *temperature_input;
    std::vector<int64_t> temperature_input_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&temperature_input, "temperature_input",
                                       temperature_input_shape, "FP32"),
                "unable to create 'temperature_input'");
    std::shared_ptr<tc::InferInput> temperature_input_ptr(temperature_input);
    FAIL_IF_ERR(temperature_input_ptr->Reset(),
                "unable to reset 'temperature_input'");
    FAIL_IF_ERR(temperature_input_ptr->AppendRaw(
                    reinterpret_cast<uint8_t *>(&config.temperature_input_data),
                    sizeof(int32_t)),
                "unable to set data for 'temperature_input'");
    std::cout << "nlp_param->temperature_input: "
              << config.temperature_input_data << std::endl;

    // input: top_k_input
    int32_t top_k_input_data = 40;
    tc::InferInput *top_k_input;
    std::vector<int64_t> top_k_input_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&top_k_input, "top_k_input",
                                       top_k_input_shape, "INT32"),
                "unable to create 'top_k_input'");
    std::shared_ptr<tc::InferInput> top_k_input_ptr(top_k_input);
    FAIL_IF_ERR(top_k_input_ptr->Reset(), "unable to reset 'top_k_input'");
    FAIL_IF_ERR(top_k_input_ptr->AppendRaw(
                    reinterpret_cast<uint8_t *>(&config.top_k_input_data),
                    sizeof(int32_t)),
                "unable to set data for 'top_k_input'");
    std::cout << "nlp_param->top_k_input: " << config.top_k_input_data
              << std::endl;

    // input: top_p_input
    float top_p_input_data = 0.95;
    tc::InferInput *top_p_input;
    std::vector<int64_t> top_p_input_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&top_p_input, "top_p_input",
                                       top_p_input_shape, "FP32"),
                "unable to create 'top_p_input'");
    std::shared_ptr<tc::InferInput> top_p_input_ptr(top_p_input);
    FAIL_IF_ERR(top_p_input_ptr->Reset(), "unable to reset 'top_p_input'");
    FAIL_IF_ERR(top_p_input_ptr->AppendRaw(
                    reinterpret_cast<uint8_t *>(&config.top_p_input_data),
                    sizeof(int32_t)),
                "unable to set data for 'top_p_input'");
    std::cout << "nlp_param->top_p_input: " << config.top_p_input_data
              << std::endl;

    // input: n_keep_input
    // TODO n_keep
    int32_t n_keep_input_data = 0;
    tc::InferInput *n_keep_input;
    std::vector<int64_t> n_keep_input_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&n_keep_input, "n_keep_input",
                                       n_keep_input_shape, "INT32"),
                "unable to create 'n_keep_input'");
    std::shared_ptr<tc::InferInput> n_keep_input_ptr(n_keep_input);
    FAIL_IF_ERR(n_keep_input_ptr->Reset(), "unable to reset 'n_keep_input'");
    FAIL_IF_ERR(n_keep_input_ptr->AppendRaw(
                    reinterpret_cast<uint8_t *>(&config.n_keep_input_data),
                    sizeof(int32_t)),
                "unable to set data for 'n_keep_input'");
    std::cout << "nlp_param->n_keep_input: " << config.n_keep_input_data
              << std::endl;

    // input: n_predict_input 256
    int32_t n_predict_input_data = 512;
    tc::InferInput *n_predict_input;
    std::vector<int64_t> n_predict_input_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&n_predict_input, "n_predict_input",
                                       n_predict_input_shape, "INT32"),
                "unable to create 'n_predict_input'");
    std::shared_ptr<tc::InferInput> n_predict_input_ptr(n_predict_input);
    FAIL_IF_ERR(n_predict_input_ptr->Reset(),
                "unable to reset 'n_predict_input'");
    FAIL_IF_ERR(n_predict_input_ptr->AppendRaw(
                    reinterpret_cast<uint8_t *>(&config.n_predict_input_data),
                    sizeof(int32_t)),
                "unable to set data for 'n_predict_input'");
    std::cout << "nlp_param->n_predict_input: " << config.n_predict_input_data
              << std::endl;

    // input: cache_prompt_input True
    bool cache_prompt_input_data = true;
    tc::InferInput *cache_prompt_input;
    std::vector<int64_t> cache_prompt_input_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&cache_prompt_input,
                                       "cache_prompt_input",
                                       cache_prompt_input_shape, "BOOL"),
                "unable to create 'cache_prompt_input'");
    std::shared_ptr<tc::InferInput> cache_prompt_input_ptr(cache_prompt_input);
    FAIL_IF_ERR(cache_prompt_input_ptr->Reset(),
                "unable to reset 'cache_prompt_input'");
    FAIL_IF_ERR(
        cache_prompt_input_ptr->AppendRaw(
            reinterpret_cast<uint8_t *>(&config.cache_prompt_input_data),
            sizeof(bool)),
        "unable to set data for 'cache_prompt_input'");
    std::cout << "nlp_param->cache_prompt_input: "
              << config.cache_prompt_input_data << std::endl;

    // input: stop_input "\n### Human"
    std::vector<std::string> stop_input_data(1);
    std::vector<int64_t> stop_input_shape{1};
    stop_input_data[0] = "<|im_end|>";
    tc::InferInput *stop_input;
    FAIL_IF_ERR(tc::InferInput::Create(&stop_input, "stop_input",
                                       stop_input_shape, "BYTES"),
                "unable to get stop_input");
    std::shared_ptr<tc::InferInput> stop_input_ptr;
    stop_input_ptr.reset(stop_input);
    stop_input_ptr->SetBinaryData(true);
    FAIL_IF_ERR(stop_input_ptr->AppendFromString(config.stop_input_data),
                "unable to set data for stop_input");

    std::cout << "nlp_param->stop_input: " << config.stop_input_data[0]
              << std::endl;

    // input: lora_scale_input [0.0, 1.0]

    std::vector<float> lora_scale_input_data = {0, 0};
    // std::cout << lora_scale_input_data << std::endl;
    // int32_t lora_scale_input_data = 0.9;
    tc::InferInput *lora_scale_input;
    std::vector<int64_t> lora_scale_input_shape{2};
    FAIL_IF_ERR(tc::InferInput::Create(&lora_scale_input, "lora_scale_input",
                                       lora_scale_input_shape, "FP32"),
                "unable to create 'lora_scale_input'");
    std::shared_ptr<tc::InferInput> lora_scale_input_ptr(lora_scale_input);
    FAIL_IF_ERR(lora_scale_input_ptr->Reset(),
                "unable to reset 'lora_scale_input'");
    FAIL_IF_ERR(
        lora_scale_input_ptr->AppendRaw(
            reinterpret_cast<uint8_t *>(config.lora_scale_input_data.data()),
            sizeof(int32_t) * 2),
        "unable to set data for 'lora_scale_input'");
    for (float data : config.lora_scale_input_data) {
        std::cout << "nlp_param->lora_scale_input: " << data << std::endl;
    }

    // input: stream_input True
    bool stream_input_data = true;
    tc::InferInput *stream_input;
    std::vector<int64_t> stream_input_shape{1};
    FAIL_IF_ERR(tc::InferInput::Create(&stream_input, "stream_input",
                                       stream_input_shape, "BOOL"),
                "unable to create 'stream_input'");
    std::shared_ptr<tc::InferInput> stream_input_ptr(stream_input);
    FAIL_IF_ERR(stream_input_ptr->Reset(), "unable to reset 'stream_input'");
    FAIL_IF_ERR(stream_input_ptr->AppendRaw(
                    reinterpret_cast<uint8_t *>(&config.stream_input_data),
                    sizeof(bool)),
                "unable to set data for 'stream_input'");

    // std::vector<tc::InferInput*> inputs = {request_type_ptr.get(),
    // text_input_ptr.get(), temperature_input_ptr.get(),
    //                                       top_k_input_ptr.get(),
    //                                       top_p_input_ptr.get(),
    //                                       n_keep_input_ptr.get(),
    //                                       n_predict_input_ptr.get(),
    //                                       cache_prompt_input_ptr.get(),
    //                                       stop_input_ptr.get(),
    //                                       stream_input_ptr.get(),
    //                                       lora_scale_input_ptr.get()};
    inputs.push_back(request_type_ptr.get());
    inputs.push_back(request_id_ptr.get());
    inputs.push_back(text_input_ptr.get());
    inputs.push_back(temperature_input_ptr.get());
    inputs.push_back(top_k_input_ptr.get());
    inputs.push_back(top_p_input_ptr.get());
    inputs.push_back(n_keep_input_ptr.get());
    inputs.push_back(n_predict_input_ptr.get());
    inputs.push_back(cache_prompt_input_ptr.get());
    inputs.push_back(stop_input_ptr.get());
    inputs.push_back(stream_input_ptr.get());
    inputs.push_back(lora_scale_input_ptr.get());
    for (auto input : inputs) {
        std::cout << "input->name: " << input->Name() << std::endl;
    }

    tc::InferRequestedOutput *text_output;
    FAIL_IF_ERR(tc::InferRequestedOutput::Create(&text_output, "text_output"),
                "unable to get text_output");
    std::shared_ptr<tc::InferRequestedOutput> text_output_ptr;
    text_output_ptr.reset(text_output);

    std::vector<const tc::InferRequestedOutput *> outputs = {
        text_output_ptr.get()};
    tc::InferOptions options(modelName_);
    options.model_version_ = "";
    tc::Error err_dsyncstreaminfer =
        grpcClient_->AsyncStreamInfer(options, inputs, outputs);
    if (!err_dsyncstreaminfer.IsOk()) {
        std::cout
            << __LINE__
            << " unable to create grpc client, err_dsyncstreaminfer.Message(): "
            << err_dsyncstreaminfer.Message() << std::endl;
        std::string end_flag = "";
    }
}

std::string LLM::chat(const std::string &prompt, const int &task_id) {
    bool verbose = true;
    // FAIL_IF_ERR(tc::InferenceServerGrpcClient::Create(&grpcClient_,
    // grpcServerUrl_, verbose), "unable to create grpc client");

    // tc::Error err_create =
    // tc::InferenceServerGrpcClient::Create(&grpcClient_, grpcServerUrl_,
    // verbose);
    result_list_.clear();
    struct chatConfig config;

    int lora_count = getLoraCount();
    std::vector<float> lora_scale_input_data(lora_count, 0.0f);
    setLoraScale(task_id, lora_scale_input_data);

    config.request_type_data = 0;
    config.request_id = request_id_;
    config.chatPrompt = prompt;
    config.temperature_input = 0.0;
    config.temperature_input_data = 0.0;
    config.top_k_input_data = 40;
    config.top_p_input_data = 0.95;
    config.n_keep_input_data = 0;
    config.n_predict_input_data = n_predict_input_;
    config.cache_prompt_input_data = true;
    config.stop_input_data = {"<|im_end|>"};
    config.lora_scale_input_data = lora_scale_input_data;
    config.stream_input_data = true;
    std::vector<tc::InferInput *> inputs;
    // tc::Headers http_headers;
    // uint32_t stream_timeout = 0;
    std::string ans;
    bool finalResponse = false;

    loadClient_->LoadModel(modelName_);
    chat_inputs(config, inputs);
    {
        auto now = std::chrono::system_clock::now();
        auto value = now.time_since_epoch();
        auto microseconds =
            std::chrono::duration_cast<std::chrono::microseconds>(value)
                .count();
        double seconds = microseconds / 1000000.0;
        printf("%d: after infer request:%lf\n", __LINE__, seconds);
    }
    int recv_count = 0;
    while (!finalResponse) {
        // Wait until all callbacks are invoked
        {
            std::unique_lock<std::mutex> lk(mutex_);
            cv_.wait(lk, [&]() {
                if (result_list_.size() > 0) {
                    return true;
                } else {
                    return false;
                }
            });
            std::shared_ptr<tc::InferResult> result = result_list_.front();
            result->IsFinalResponse(&finalResponse);
            std::vector<std::string> text_output_data;
            result->StringData("text_output", &text_output_data);

            if (text_output_data.size() != 1) {
                std::cerr << "error: received incorrect number of strings for "
                             "text_output: "
                          << text_output_data.size() << std::endl;
                break;
                // return ans;
            }
            {
                auto now = std::chrono::system_clock::now();
                auto value = now.time_since_epoch();
                auto microseconds =
                    std::chrono::duration_cast<std::chrono::microseconds>(value)
                        .count();
                double seconds = microseconds / 1000000.0;
                printf("%d: get token output:%lf\n", __LINE__, seconds);
            }
            ans += text_output_data[0];
            printf("prompt:%s\n<----->nlp engine get from triton llm:%s\n",
                   prompt.c_str(), ans.c_str());
            completionCallbacks_(text_output_data[0], false);
            result_list_.erase(result_list_.begin());
        }
    }
    std::string end_flag = "";
    completionCallbacks_(end_flag, true);
    // grpcClient_->StopStream();

    return ans;
}
