/*
 * Copyright 2024 KylinSoft Co., Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <https://www.gnu.org/licenses/>.
 */

#include "ondevicenlpengine.h"
#include "nlp/llm.h"

#include <jsoncpp/json/json.h>
#include <nlohmann/json.hpp>

namespace ai_engine::gen_ai::nlp {
OndeviceNlpEngine::OndeviceNlpEngine() : llm_(new LLM()) {}

OndeviceNlpEngine::~OndeviceNlpEngine() {
    result::AiEngineBaseResult error;
    error = destroy();
};

void OndeviceNlpEngine::setTextGenerationResultCallback(
    nlp::TextGenerationResultCallback callback) {
    chatResultCallback_ = std::move(callback);

    if (llm_ == nullptr) {
        return;
    }

    auto chatCallback = [this](const std::string &result, bool over) {
        this->onChatResult(result, over);
    };
    llm_->setCompletionResultCallback(chatCallback);
}

result::AiEngineBaseResult OndeviceNlpEngine::chat(const std::string &message) {
    result::AiEngineBaseResult error;
    error.success = true;
    chatStopped_ = false;

    if (llm_ == nullptr) {
        error.success = false;
        error.error = ai_engine::EngineError(
            (int)NlpEngineErrorCode::FailedToConnectServer,
            "[OndeviceNlpEngine]: Nlp is nullptr");
        return error;
    }
    if (!inited_) {
        error = init();
        if (!error.success) {
            return error;
        }
        setTextGenerationResultCallback(chatResultCallback_);
    }
    Json::Reader reader;
    Json::Value bodys;
    if (!reader.parse(message, bodys)) {
        error.success = false;
        error.error = ai_engine::EngineError(
            (int)NlpEngineErrorCode::FailedToConnectServer,
            "[OndeviceNlpEngine]: Nlp message parse error");
        return error;
    }
    std::string qus = "\n<|im_start|>";
    for (auto body : bodys["messages"]) {
        qus += body["role"].asString();
        qus += "\n";
        qus += body["content"].asString();
        qus += "<|im_end|>\n<|im_start|>";
    }
    qus += "assistant\n";
    int task_id = -1;
    if (!bodys["task_id"].isNull()) {
        task_id = bodys["task_id"].asInt();
        std::cout << "bodys task_id not null" << std::endl;
    }
    std::cout << qus << std::endl;
    std::cout << "task_id:" << task_id << std::endl;

    llm_->chat(qus, task_id);

    return error;
}

void OndeviceNlpEngine::stopChat() {
    chatStopped_ = true;

    TextGenerationResult endResult;
    endResult.success = true;
    endResult.error = ai_engine::EngineError();
    endResult.modelName = llm_->modelName_;
    endResult.text =
        std::string{"{\"result\":\"<|stopChat_8167431|>\",\"is_end\":true}"};

    if (llm_ == nullptr) {
        return;
    }
    llm_->stopChat();
    if (chatResultCallback_ != nullptr) {
        chatResultCallback_(endResult);
    }
}

void OndeviceNlpEngine::onChatResult(const std::string &result, bool over) {
    if (chatResultCallback_ == nullptr) {
        return;
    }

    if (chatStopped_) {
        return;
    }

    Json::Value body;
    body["is_end"] = over;
    body["result"] = result;

    TextGenerationResult res;
    res.success = true;
    res.error = ai_engine::EngineError();
    res.modelName = llm_->modelName_;
    res.text = body.toStyledString();
    chatResultCallback_(res);
}

result::AiEngineBaseResult OndeviceNlpEngine::init() {
    result::AiEngineBaseResult error;

    error.success = true;
    if (inited_) {
        error.success = false;
        error.error = ai_engine::EngineError(
            (int)NlpEngineErrorCode::FilePermissionDenied,
            "[OndeviceNlpEngine]: Nlp early inited");
        return error;
    }
    bool ret = llm_->llm_load_model();
    if (!ret) {
        error.success = false;
        error.error = ai_engine::EngineError(
            (int)NlpEngineErrorCode::FilePermissionDenied,
            "[OndeviceNlpEngine]: Nlp model load error");
        return error;
    }

    inited_ = true;
    return error;
}

result::AiEngineBaseResult OndeviceNlpEngine::destroy() {
    result::AiEngineBaseResult error;
    error.success = true;
    if (!inited_) {
        return error;
    }
    inited_ = false;
    releaseTimerCondition_.notify_all();

    delete llm_;
    llm_ = nullptr;
    return error;
}

namespace nl = nlohmann;

bool OndeviceNlpEngine::setCurrentModel(const std::string &model) {
    bool res = false;
    // if(model.size() > 0 && llm_->setLlmModelName(model)) {
    //     res = true;
    // } else {
    //     res = false;
    // }

    return res;
}

std::string OndeviceNlpEngine::getLocalModelList() const {
    const char *info = R"(
    {
        "vendor": "",
        "models": [
        ]
    }
    )";
    nl::json llm_json;
    std::string main_info(info);
    nl::json main_json = nl::json::parse(main_info);
    std::string model_info_llm = llm_->GetConfigValue("model_info_list", "llm");
    printf("model_info_llm:%s\n", model_info_llm.c_str());
    if (model_info_llm.size() > 0)
        llm_json = nl::json::parse(model_info_llm);
    else
        llm_json = nl::json::parse(main_info);

    main_json["models"].insert(main_json["models"].end(),
                               llm_json["models"].begin(),
                               llm_json["models"].end());
    return main_json.dump(4);
}

std::string OndeviceNlpEngine::currentModel() const {
    return llm_->modelName_;
};

std::string OndeviceNlpEngine::modelInfo() const {
    // llm_->modelInfoUpdate();
    printf("modelInfo()---modelInfo_:%s\n", llm_->modelInfo_.c_str());
    return llm_->modelInfo_;
}

} // namespace ai_engine::gen_ai::nlp
